"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CodeMirror = require("codemirror");
const lsp_editor_adapter_1 = require("lsp-editor-adapter");
const utils_1 = require("../utils");
const converter_1 = require("../converter");
const lsp_1 = require("../lsp");
// TODO: settings
const hover_modifier = 'Control';
class CodeMirrorAdapterExtension extends lsp_editor_adapter_1.CodeMirrorAdapter {
    constructor(connection, options, editor, create_tooltip, invoke_completer) {
        super(connection, options, editor);
        this.marked_diagnostics = new Map();
        this.create_tooltip = create_tooltip;
        this.invoke_completer = invoke_completer;
        // @ts-ignore
        let listeners = this.editorListeners;
        let wrapper = this.editor.getWrapperElement();
        // detach the adapters contextmenu
        wrapper.removeEventListener('contextmenu', listeners.contextmenu);
        // TODO: actually we only need the connection...
        //  the tooltips and suggestions will need re-writing to JL standards anyway
        // show hover after pressing the modifier key
        wrapper.addEventListener('keydown', (event) => {
            if ((!hover_modifier || utils_1.getModifierState(event, hover_modifier)) &&
                this.hover_character === this.last_hover_character) {
                this.show_next_tooltip = true;
                this.handleHover(this.last_hover_response);
            }
        });
        wrapper.addEventListener('keyup', () => {
            // TODO: the updates frequency and triggers will require a review and a clean up
            this.connection.sendChange();
        });
    }
    get hover_character() {
        // @ts-ignore
        return this.hoverCharacter;
    }
    handleGoTo(locations) {
        this.remove_tooltip();
        // do NOT handle GoTo actions here
    }
    handleCompletion(completions) {
        // do NOT handle completion here
        // TODO: UNLESS the trigger character was typed!
    }
    static get_markup_for_hover(response) {
        let contents = response.contents;
        // this causes the webpack to fail "Module not found: Error: Can't resolve 'net'" for some reason
        // if (lsProtocol.MarkedString.is(contents))
        ///  contents = [contents];
        if (typeof contents === 'string') {
            contents = [contents];
        }
        if (!Array.isArray(contents)) {
            return contents;
        }
        // now we have MarkedString
        let content = contents[0];
        if (typeof content === 'string') {
            // coerce to MarkedString  object
            return {
                kind: 'plaintext',
                value: content
            };
        }
        else {
            return {
                kind: 'markdown',
                value: '```' + content.language + '\n' + content.value + '```'
            };
        }
    }
    remove_tooltip() {
        // @ts-ignore
        this._removeHover(); // this removes underlines
        if (this._tooltip !== undefined) {
            this._tooltip.dispose();
        }
    }
    handleHover(response) {
        this.remove_tooltip();
        if (!response ||
            !response.contents ||
            (Array.isArray(response.contents) && response.contents.length === 0)) {
            return;
        }
        this.highlight_range(response.range, 'cm-lp-hover-available');
        this.last_hover_response = null;
        if (!this.show_next_tooltip) {
            this.last_hover_response = response;
            this.last_hover_character = this.hover_character;
            return;
        }
        const markup = CodeMirrorAdapterExtension.get_markup_for_hover(response);
        let position = this.hover_character;
        let cm_editor = this.get_cm_editor(position);
        this._tooltip = this.create_tooltip(markup, cm_editor, position);
    }
    get_cm_editor(position) {
        // TODO necessity to have dependency on position is where the idea of mapping notebooks
        //  with an object pretending to be an editor has a weak side...
        let cm_editor = this.editor;
        if (cm_editor.get_editor_at !== undefined) {
            cm_editor = cm_editor.get_editor_at(position);
        }
        return cm_editor;
    }
    get_language_at(position, editor) {
        if (typeof editor === 'undefined') {
            editor = this.editor;
        }
        return editor.getModeAt(position).name;
    }
    get_markup_for_signature_help(response, language) {
        let signatures = new Array();
        response.signatures.forEach((item) => {
            let markdown = '```' + language + '\n' + item.label + '\n```';
            if (item.documentation) {
                markdown += '\n';
                // TODO: make use of the MarkupContent object instead
                for (let line of item.documentation.toString().split('\n')) {
                    if (line.trim() == item.label.trim()) {
                        continue;
                    }
                    if (line.startsWith('>>>')) {
                        line = '```' + language + '\n' + line.substr(3) + '\n```';
                    }
                    markdown += line + '\n';
                }
            }
            signatures.push(markdown);
        });
        return {
            kind: 'markdown',
            value: signatures.join('\n\n')
        };
    }
    handleSignature(response) {
        this.remove_tooltip();
        // @ts-ignore
        let token = this.token;
        if (!token || !response || !response.signatures.length) {
            return;
        }
        let position = token.start;
        let language = this.get_language_at(position);
        let markup = this.get_markup_for_signature_help(response, language);
        let cm_editor = this.get_cm_editor(position);
        this._tooltip = this.create_tooltip(markup, cm_editor, position);
    }
    handleChange(cm, change) {
        // based on https://github.com/wylieconlon/lsp-editor-adapter/blob/e80e44310f8c12f87f3da9be07772102610ce517/src/codemirror-adapter.ts#L65
        // ISC licence (TODO: move this to the fork to separate out the ISC code)
        this.remove_tooltip();
        const location = this.editor.getDoc().getCursor('end');
        this.connection.sendChange();
        const completionCharacters = this.connection.getLanguageCompletionCharacters();
        const signatureCharacters = this.connection.getLanguageSignatureCharacters();
        const code = this.editor.getValue();
        const lines = code.split('\n');
        const line = lines[location.line];
        const typedCharacter = line[location.ch - 1];
        if (completionCharacters.indexOf(typedCharacter) > -1) {
            this.invoke_completer();
        }
        else if (signatureCharacters.indexOf(typedCharacter) > -1) {
            // @ts-ignore
            this.token = this._getTokenEndingAtPosition(code, location, signatureCharacters);
            this.connection.getSignatureHelp(location);
        }
    }
    highlight_range(range, class_name) {
        let hover_character = this.hover_character;
        let start;
        let end;
        if (range) {
            start = converter_1.PositionConverter.lsp_to_cm(range.start);
            end = converter_1.PositionConverter.lsp_to_cm(range.end);
        }
        else {
            // construct range manually using the token information
            let token = this.editor.getTokenAt(hover_character);
            start = { line: hover_character.line, ch: token.start };
            end = { line: hover_character.line, ch: token.end };
        }
        // @ts-ignore
        this.hoverMarker = this.editor.getDoc().markText(start, end, {
            className: class_name
        });
    }
    handleMouseOver(event) {
        // proceed when no hover modifier or hover modifier pressed
        this.show_next_tooltip =
            !hover_modifier || utils_1.getModifierState(event, hover_modifier);
        return super.handleMouseOver(event);
    }
    // duck typing: to enable use of notebook mapper
    get transform() {
        const notebook_as_editor = this.editor;
        if (notebook_as_editor.transform !== undefined) {
            return position => notebook_as_editor.transform(position);
        }
        else {
            return position => position;
        }
    }
    get_editor_index(position) {
        const notebook_as_editor = this.editor;
        if (notebook_as_editor.get_cell_at !== undefined) {
            let cell = notebook_as_editor.get_cell_at(position);
            return notebook_as_editor.notebook.widgets.findIndex(other_cell => {
                return cell == other_cell;
            });
        }
        else {
            return 0;
        }
    }
    get get_cell_id() {
        const notebook_as_editor = this.editor;
        if (notebook_as_editor.get_cell_at !== undefined) {
            return position => notebook_as_editor.get_cell_at(position).id;
        }
        else {
            return position => '';
        }
    }
    handleDiagnostic(response) {
        /*
        TODO: the base class has the gutter support, like this
        this.editor.clearGutter('CodeMirror-lsp');
         */
        // Note: no deep equal for Sets or Maps in JS:
        // https://stackoverflow.com/a/29759699
        const markers_to_retain = new Set();
        // add new markers, keep track of the added ones
        let doc = this.editor.getDoc();
        let transform = this.transform;
        let get_cell_id = this.get_cell_id;
        response.diagnostics.forEach((diagnostic) => {
            const start = converter_1.PositionConverter.lsp_to_cm(diagnostic.range.start);
            const end = converter_1.PositionConverter.lsp_to_cm(diagnostic.range.end);
            const severity = lsp_1.diagnosticSeverityNames[diagnostic.severity];
            // what a pity there is no hash in the standard library...
            // we could use this: https://stackoverflow.com/a/7616484 though it may not be worth it:
            //   the stringified diagnostic objects are only about 100-200 JS characters anyway,
            //   depending on the message length; this could be reduced using some structure-aware
            //   stringifier; such a stringifier could also prevent the possibility of having a false
            //   negative due to a different ordering of keys
            // obviously, the hash would prevent recovery of info from the key.
            let diagnostic_hash = JSON.stringify(Object.assign({}, diagnostic, { 
                // the apparent marker position will change in the notebook with every line change for each marker
                // after the (inserted/removed) line - but such markers should not be invalidated,
                // i.e. the invalidation should be performed in the cell space, not in the notebook coordinate space,
                // thus we transform the coordinates and keep the cell id in the hash
                range: {
                    start: transform(start),
                    end: transform(end)
                }, cell: get_cell_id(start) }));
            markers_to_retain.add(diagnostic_hash);
            if (!this.marked_diagnostics.has(diagnostic_hash)) {
                let options = {
                    title: diagnostic.message +
                        (diagnostic.source ? ' (' + diagnostic.source + ')' : ''),
                    className: 'cm-lsp-diagnostic cm-lsp-diagnostic-' + severity
                };
                let marker;
                try {
                    marker = doc.markText(start, end, options);
                }
                catch (e) {
                    console.warn('Marking inspection (diagnostic text) failed, see following logs (2):');
                    console.log(diagnostic);
                    console.log(e);
                    return;
                }
                this.marked_diagnostics.set(diagnostic_hash, marker);
            }
            /*
            TODO and this:
              const childEl = document.createElement('div');
              childEl.classList.add('CodeMirror-lsp-guttermarker');
              childEl.title = diagnostic.message;
              this.editor.setGutterMarker(start.line, 'CodeMirror-lsp', childEl);
            do we want gutters?
           */
        });
        // remove the markers which were not included in the new message
        this.marked_diagnostics.forEach((marker, diagnostic_hash) => {
            if (!markers_to_retain.has(diagnostic_hash)) {
                this.marked_diagnostics.delete(diagnostic_hash);
                marker.clear();
            }
        });
    }
}
exports.CodeMirrorAdapterExtension = CodeMirrorAdapterExtension;
